// comm.cpp
//
//   Protocol.
//
#include <windows.h>
//#include <iostream.h>
#include <stdio.h>
#include <stdlib.h>
#include <conio.h>
#include "general.h"
#include "comm.h"

extern int sendpacket(unsigned char *pc);
int debug = 0;

char *command_name[] = {
	"OK",
	"DENY",
	"NEWGAME",
	"QUERY",
	"ANSWER",
	"MOVE",
	"TAKEBACK",
	"EXTEND"
};
int querylist[] = {
	Q_COLOR, Q_HANDICAP, -1};

Port::Port (void)
{
	my_sequence = other_sequence = 0;
	is_running = FALSE;
	my_used_time = others_used_time = 0;
	handicap = 0;
	board_size = 19;
	my_color = STONE_BLACK;
	recieved = FALSE;
	index = 0;
	nextquery = 0;
	state = S_NONE;
}

// ------------------------------------------------------------------
// Port::SetCommand --
//   Set the command to send to the other.
//   1. Calculate bytes from the command No. and an argument.
//   2. Set the sequence number.
//   3. Calculate the checksum.
//
// Args:
//   com = command No.
//   arg = arthment of the command.
// ------------------------------------------------------------------
void Port::SetCommand (int com, int arg)
{
	if (com == P_OK) arg = 0x3ff;
	arg &= 0x3ff;
	ch[2] = (com << 4) | ((arg >> 7) & 0x07);
	ch[2] |= 0x80;
	ch[2] &= 0xf7;			// bit 4 must be 0.
	ch[3] = (arg & 0x7f) | 0x80;

	// Add sequence number.
	if (com != P_OK) {
		my_sequence = 1 - my_sequence;	// 0 or 1.
	}

	ch[0] = my_sequence | (other_sequence << 1);
	ch[1] = get_checksum (ch);
}

void Port::PutQuery(){

	query_code = querylist[nextquery];
	if(query_code != -1) {
		waiting = TRUE;
		Send(P_QUERY,query_code);
	} else {
		Send(P_OK,0x3ff);
		state = S_GAME;
	}
}

// ------------------------------------------------------------------
// Port::get_checksum --
//   Return the checksum of the argument.
//
// Args:
//   Data buffer to be send.
//
// Return:
//   uchar = the checksum.
// ------------------------------------------------------------------
unsigned char Port::get_checksum (unsigned char *c)
{
	return ((c[0] + c[2] + c[3]) & 0x7f) | 0x80;
}

// ------------------------------------------------------------------
// Port::get_command --
//   Get the command No. from data.
//
// Args:
//   none
//
// Return:
//    int = command No.
// ------------------------------------------------------------------
int Port::get_command (void)
{
	return (ch[2] >> 4) & 0x7;
}

// ------------------------------------------------------------------
// Port::get_arg --
//   Get the argument from current buffer.
//
// Args: none
//
// Return:
//   int = The argument.
// ------------------------------------------------------------------
int Port::get_arg (void)
{
	return ((ch[2] & 0x07) << 7) | (ch[3] & 0x7f);
}

// ------------------------------------------------------------------
// Port::Parse --
//   Read the packect from the buffer and parse it.
//
// Args:
// ------------------------------------------------------------------
void Port::Parse (BYTE c)
{
	if(index == 0) {
		if((c & 0xfc) == 0) { //find start byte
			ch[0] = c;
			index = 1;
		}
	}
	else {
		if(( c & 0x80) != 0) {
			ch[index] = c;
			index++;
			if(index == 4) {
				index = 0;
				exec();
			}
		} else if( c & 0xfc == 0){
			ch[0] = c;
			index =	1;
		}
	}
}

// ------------------------------------------------------------------
// Port::exec --
//   Execute the read data.
//   1. check the checksum
//   2. check the sequence number.
//   3. parse and execute the command.
// ------------------------------------------------------------------
void Port::exec (void)
{
	if (get_checksum (ch) != ch[1]) {
		goerror ("Wrong checksum.");
		return;
	}
	command = get_command ();
	arg = get_arg ();
	if (debug) {
		cprintf ("Receive: com=%d arg=%d(%2x %2x %2x %2x)\r\n", command, arg,
			ch[0], ch[1], ch[2], ch[3]);
	}
	// Check the sequence number.
	//  1byte - 0000 00hy
	//     h = his sequence, it is my sequence, now.
	//     y = your sequence, it is other's sequence, now.
	int other_seq = ch[0] & 1;
	int my_seq = (ch[0] & 2) >> 1;

	// Check sequence number.
	if (waiting == TRUE) {
		if (command == P_OK) {
			if (my_sequence != my_seq) { 
				gostatus ("Warn: OK - My sequence number is old.");
				return;
			} else if (other_seq != other_sequence) {
				gostatus ("Warn: OK - His sequence is not equal.");
				return;
			}
		} else {
			if (my_sequence != my_seq && other_sequence != other_seq) {
				gostatus ("Warn: CMD - Conflict.");
				// Revert my sequence id and go to neutral.
				my_sequence = 1 - my_sequence;
//				state = S_GAME;
				return;
			} else if (my_sequence != my_seq && other_sequence == other_seq) {
				gostatus ("Warn: CMD - His command is old.");
				SendAgain ();
				return;
			} else if (my_sequence == my_seq && other_sequence != other_seq) {
				// He must have sent OK at that time.  Accept this command.
			} else {
				gostatus ("Warn: CMD - Not possible, discard.");
				return;
			}
		}
	} else { 	// not waiting OK.
		if (command == P_OK) {
			gostatus ("Warn: OK - not waiting, discard.");
			return;
		} else if (my_sequence != my_seq) {
			gostatus ("Warn: CMD - My sequence is not updated.");
			return;
		} else if (other_sequence == other_seq) {
			gostatus ("Warn: CMD - His command is old. Send OK again.");
			Send (P_OK, 0x3ff);
			return;
		}
	}

	other_sequence = other_seq;
	doCommand (command, arg);	// Parse and execute the command.
}

// ------------------------------------------------------------------
// Port::doCommand --
//   Execute the command.
// ------------------------------------------------------------------
void Port::doCommand (int com, int arg)
{
	switch (com) {
	case P_OK:
		if (state == S_NEW) state = S_GAME;
		if (state == S_MOVE) {// Send 'MOVE' and received the ACK.
			state = S_GAME;
		}
		waiting = FALSE;
		break;

	case P_DENY:
		if (state == S_NEW) {
			// NEW command is not accepted.
			state = S_NONE;
			return;
		} else if (state == S_MOVE) {	// MOVE command is denied.
			state = S_GAME;
		}
		waiting = FALSE;
		break;

	case P_NEWGAME:
		state = S_NEW;
		nextquery = 0;
		PutQuery();
		break;

	case P_QUERY:
		switch (arg) {
		case Q_GAME:	// Answer the game "Go"
			Send (P_ANSWER, 1);
			break;
		case Q_BUFFER_SIZE:
			Send (P_ANSWER, 0);	// return minimum size (4 bytes).
			break;
		case Q_VERSION:
			Send (P_ANSWER, 0);	// Protocol version, return 0.
			break;
		case Q_STONES:
			Send (P_ANSWER, blackCount + whiteCount);
			// Return the number of stone on the board.
			break;
		case Q_BLACK_TIME:	// Black's used time
			if (my_color == STONE_BLACK) {
				Send (P_ANSWER, my_used_time);
			} else {
				Send (P_ANSWER, others_used_time);
			}
			break;
		case Q_WHITE_TIME:	// White's used time.
			if (my_color == STONE_WHITE) {
				Send (P_ANSWER, my_used_time);
			} else {
				Send (P_ANSWER, others_used_time);
			}
			break;
		case Q_CHAR:	// Character code.
			Send (P_ANSWER, 1);	// ASCII only
			break;
		case Q_RULE:	// Kind of rule.
			Send (P_ANSWER, 1);	// Japanese rule
			break;
		case Q_HANDICAP:	// Handicap
			Send (P_ANSWER, 1);	// No handicap
			break;
		case Q_BOARD_SIZE:	// Board size
			Send (P_ANSWER, 19);	// 19x19
			break;
		case Q_TIMELIMIT:
			Send (P_ANSWER, 0);	// unknown
			break;
		case Q_COLOR:    // My color
			if (my_color == STONE_BLACK) {
				Send (P_ANSWER, 2);
			} else {
				Send (P_ANSWER, 1);
			}
			break;
		case Q_WHO:	// Who am I?
			Send (P_ANSWER, 0);	// unknown.
			break;
		default:
			Send (P_ANSWER, 0);	// unknown.
		}
		break;

	case P_ANSWER:
		switch (query_code) {
		case Q_HANDICAP:	// Handicap
				break;
		case Q_COLOR:
				break;
		}
		nextquery++;
		PutQuery();
		break;

	case P_MOVE:
		int stone = (arg & 0x200) ? STONE_WHITE : STONE_BLACK;
		if (state == S_GAME) {
			if (stone != my_color) {
				Send (P_OK, 0x3FF);
				ParseMove (arg);
			} else {
				goerror ("You send wrong color.");
				Send (P_DENY, 0);
			}
		}
		break;

/* case P_TAKEBACK:
	goerror ("Takeback is not supported.");
	Send (P_DENY, 0);
	break;

  case P_EXTEND:
	goerror ("Extend command is not supported.");
	Send (P_DENY, 0);
	break;

  default:
	goerror ("Wrong command No. is sent.");
*/
	}
}

// ------------------------------------------------------------------
// Port::ParseMove --
//    the move sent from other.
//
// Args:
//   val = other's argument.
//
// ------------------------------------------------------------------
void Port::ParseMove (int val)
{
//  int color = (val & 0x200) ? STONE_WHITE : STONE_BLACK;
  val &= 0x1ff;
  if (val == 0) {
	rx = ry = 0;		// Pass
  } else {
	val--;
	rx = val % board_size + 1;
	ry = board_size - (val / board_size);
  }
  recieved = TRUE;
}

// ------------------------------------------------------------------
// Port::SetHandicap --
//   Set my handicap.
//
// Args:
//  handi = more than 0, 1(even), 2, 3, ....
//
// ------------------------------------------------------------------
void Port::SetHandicap (int handi)
{
  if (handi > 0) handi--;	// 1 = even.
  handicap = handi;
}

// ------------------------------------------------------------------
// Port::NewGame --
//   Initialize the game status.
//
// Args:
//   col = STONE_BLACK or STONE_WHITE
// ------------------------------------------------------------------
void Port::NewGame (int color)
{
	my_color = color;                // color must be 1:Black or 2:White
	if (color == STONE_BLACK) {
		waiting = TRUE;
		state = S_NEW;
		Send (P_NEWGAME, 0);
	}
}

// ------------------------------------------------------------------
// Port::Send --
//   Send the command to serial port.
// ------------------------------------------------------------------
void Port::Send (int send_command, int send_arg)
{
  SetCommand (send_command, send_arg);
  // by MAB
  memcpy(lastch,ch,sizeof(ch));
  if (debug) {
	cprintf ("Send: com=%d arg=%d (%2x %2x %2x %2x)\r\n",
	  send_command, send_arg, ch[0], ch[1], ch[2], ch[3]);
  }
  sendpacket (ch);
}

// ------------------------------------------------------------------
// Port::SendAgain --
//   Send the data again.
// ------------------------------------------------------------------
void Port::SendAgain (void)
{
	if (debug) {
	  cprintf ("Send Again: %2x %2x %2x %2x\r\n",
	lastch[0], lastch[1], lastch[2], lastch[3]);
	}
	sendpacket (lastch);	// Fixed by MAB
	// The old program used the same buffer for In/Out
	// therefore it always resend the packet received from the opponent,
	// instead of the last one he sent.
}

// ------------------------------------------------------------------
// Port::SendMove --
//
// Args:
//   px, py = coordinates. 1 ... 19, or 0 to pass.
//   color = STONE_WHITE or STONE_BLACK
//
// ------------------------------------------------------------------
void Port::SendMove (int sx, int sy)
{
  int num;
  if (sx == 0 && sy == 0) {
	num = 0;		// Pass this move.
  } else {
	num = sx + (board_size - sy) * board_size;
  }
  if (my_color == STONE_WHITE) {
	num |= 0x200;
  }
  waiting = TRUE;
  state = S_MOVE;
  Send (P_MOVE, num);
}

int Port::RecieveMove(int *px, int *py)
{
	if(recieved) {
		*px = rx; *py = ry;
		return TRUE;
	} else return FALSE;
}

void Port::ClearRecieve(void)
{
	recieved = FALSE;
}

int Port::CheckOK(void)
{
	return (state == S_GAME);
}

void gostatus(char *s)
{
//	cout << s << endl;
}
void goerror(char *s)
{
//	cout << s << endl;
}
