// cgfcom.cpp
#define STRICT
#define TIMEOUT 10000L
#include <windows.h>
#include <stdlib.h>
#include <stdio.h>
#include "cgfcom.h"
#include "comm.h"

Port *port;
HANDLE hComm, hThread;
BOOL success, bReading;
DCB dcb;
DWORD threadID;

void PrintError( LPCSTR str )
{
	LPVOID lpMessageBuffer;
	int error = GetLastError();
	FormatMessage(
		FORMAT_MESSAGE_ALLOCATE_BUFFER |
		FORMAT_MESSAGE_FROM_SYSTEM,
		NULL, error,
		MAKELANGID(LANG_NEUTRAL,SUBLANG_DEFAULT),
		(LPTSTR) &lpMessageBuffer,
		0, NULL
	);
	printf("%s:(%d) %s\n",str,error,lpMessageBuffer);
	LocalFree(lpMessageBuffer);
}

DWORD ReadThread()
{
/* Thread for event-driven reading of bytes from the COM port */
DWORD nBytesRead, dwEvent, dwError;
COMSTAT cs;
unsigned char c;

	/* generate event whenever a byte arives */
	if(!SetCommMask (hComm, EV_RXCHAR))
		PrintError("SetCommMask");
	
	/* read a byte whenever one is received */
	while (bReading) {
		/* wait for event */
		if (WaitCommEvent (hComm, &dwEvent, NULL)){
			if(!ClearCommError (hComm, &dwError, &cs))
				PrintError("ClearCommError");
			/* read all available bytes */
			do {
				if(ReadFile(hComm,&c,1,&nBytesRead,NULL)) 
					port->Parse(c);
				else
					break;
			} while(nBytesRead);
		} else {
			PrintError("WaitComEvent Error");
			break;
		}
	} /* end while (reading bytes in loop) */
	/* clean out any pending bytes in the receive buffer */
	PurgeComm(hComm, PURGE_RXCLEAR);
	return 0L;
} /* end function (ReadThread) */


BOOL __declspec(dllexport) initcom(int port_num)
{
	char portname[5];

	//protocol class open
	port = new Port;
	
	//open comport
	if(port_num == 1) strcpy(portname,"COM1");
	if(port_num == 2) strcpy(portname,"COM2");
	/* open the port here */
	if ((hComm=CreateFile (portname,
			GENERIC_READ|GENERIC_WRITE, 
			0, NULL, OPEN_EXISTING,
			FILE_ATTRIBUTE_NORMAL, NULL)) == INVALID_HANDLE_VALUE) {
		/* handle error */
		PrintError("Failed to open port");
		return FALSE;
	} /* end if (error opennig comport) */
	// set timeouts
	COMMTIMEOUTS cto = { MAXWORD, 0, 0, 5, 10 };
	DCB dcb;
	if(!SetCommTimeouts(hComm,&cto)) {
		PrintError("SetCommTimeouts failed");
		return  FALSE;
	}
	// set DCB
	memset(&dcb,0,sizeof(dcb));
	dcb.DCBlength = sizeof(dcb);
	dcb.BaudRate = 2400;
	dcb.fBinary = 1;
	dcb.fDtrControl = DTR_CONTROL_ENABLE;
	dcb.fOutxCtsFlow = 1;
	dcb.fRtsControl = DTR_CONTROL_HANDSHAKE;
	dcb.Parity = NOPARITY;
	dcb.StopBits = ONESTOPBIT;
	dcb.ByteSize = 8;
	if(!SetCommState(hComm,&dcb)){
		PrintError("SetCommState failed");
		return FALSE;
	}
//	EscapeCommFunction(hComm, SETDTR|SETRTS);

	/* create the thread for reading bytes */
	bReading = TRUE;
	if ((hThread=CreateThread (NULL /*def security */, 0 /* def stack size */, 
		(LPTHREAD_START_ROUTINE)ReadThread, NULL /* param to pass to thread */, 
			0, &threadID)) == INVALID_HANDLE_VALUE) {
		/* handle error */
		PrintError ("Error creating READ thread");
		CloseHandle (hComm);
		return FALSE;
	} /* end if (error creating read thread) */

	return TRUE;
}


BOOL sendpacket(unsigned char *packet)
{
	unsigned long write;
	   success = WriteFile(hComm, packet, 4, &write, NULL);
	   if (!success) 
			PrintError("sendc error");
	return success;
}

BOOL __declspec(dllexport) newgame(int color)
{
	DWORD StartCount;
	port->NewGame(color);
	StartCount = GetTickCount();
	do {
		if((GetTickCount()-StartCount) > TIMEOUT)
			return FALSE;
	} while(!(port->CheckOK()));
	return TRUE;
}

BOOL __declspec(dllexport) sendxy(int x, int y)
{
	DWORD StartCount;
	port->SendMove(x,y);
	StartCount = GetTickCount();
	do {
		if((GetTickCount()-StartCount) > TIMEOUT)
			return FALSE;
	}while(!(port->CheckOK()));
	return TRUE;
}

BOOL __declspec(dllexport) receivexy(int *px, int *py)
{
	BOOL ret;
	ret = port->RecieveMove(px,py);
	if(ret) port->ClearRecieve();
	return ret;
}

BOOL __declspec(dllexport) closecom(void)
{
	/* kill the read thread */
	bReading = FALSE;
	/* in case the thread is not running, resume it now */
	ResumeThread (hThread);
	/* clear all events and force WaitCommEvent to return */
	SetCommMask (hComm, 0L);
	PurgeComm(hComm, 
		PURGE_TXABORT|PURGE_RXABORT|PURGE_TXCLEAR|PURGE_RXCLEAR);
	CloseHandle (hComm);

	delete port;
	return TRUE;
}

