// cgfcom.cpp
#define STRICT
#define TIMEOUT 10000L
#include <windows.h>
#include <stdlib.h>
#include <stdio.h>
#include "cgfcom.h"
#include "comm.h"


// Default initialization (ALT)
Port		*port=0;
HANDLE		hComm=INVALID_HANDLE_VALUE, hThread=INVALID_HANDLE_VALUE;
BOOL		success=0, bReading=0;
DCB			dcb;
DWORD		threadID=0;
OVERLAPPED  overlapRead;
OVERLAPPED  overlapWrite;


void PrintError( LPCSTR str )
{
	LPVOID lpMessageBuffer;
	int error = GetLastError();
	FormatMessage(
		FORMAT_MESSAGE_ALLOCATE_BUFFER |
		FORMAT_MESSAGE_FROM_SYSTEM,
		NULL, error,
		MAKELANGID(LANG_NEUTRAL,SUBLANG_DEFAULT),
		(LPTSTR) &lpMessageBuffer,
		0, NULL
	);
	printf("%s:(%d) %s\n",str,error,lpMessageBuffer);
	LocalFree(lpMessageBuffer);
}

//------ ALT cleanups all data ( returns FALSE always ) 
BOOL  CleanUp(void)
{
	//--- kill the read thread 
	bReading = FALSE;

	// in case the thread is not running, resume it now */
	if (hThread!=INVALID_HANDLE_VALUE)	TerminateThread(hThread,0);
	//----- ALT ResumeThread doesn't work correctly for Overlapped I/O
	//ResumeThread (hThread);
	hThread=INVALID_HANDLE_VALUE;

	// clear all events and force WaitCommEvent to return */
	if (hComm!=INVALID_HANDLE_VALUE) {
		PurgeComm(hComm, PURGE_TXABORT|PURGE_RXABORT|PURGE_TXCLEAR|PURGE_RXCLEAR);
		CloseHandle (hComm);
		hComm=INVALID_HANDLE_VALUE;
	}
	
	if (overlapRead.hEvent) {
		CloseHandle( overlapRead.hEvent ) ;
		overlapRead.hEvent=0;
	}
	if (overlapWrite.hEvent) {
		CloseHandle( overlapWrite.hEvent ) ;
		overlapWrite.hEvent=0;
	}

	if (port) delete port;
	port=0;

	return FALSE;
}
//---------------------------- Overlapped I/O (ALT) ---------------------------
DWORD ReadThread()
{
	DWORD dwError;
	COMSTAT cs;

	DWORD nBytesRead;
	unsigned char c;

	if (hComm==INVALID_HANDLE_VALUE) return 0;
	if (port==0) return 0;
	do {
		if (!bReading) break;
		// try to read a byte 
		ClearCommError (hComm, &dwError, &cs);
		//if (cs.cbInQue==0) continue;

		BOOL success=ReadFile(hComm,&c,1,&nBytesRead,&overlapRead);
		if (!success) {
				if (GetLastError()==ERROR_IO_PENDING) {
					//------- wait until received -------------------
					while (!GetOverlappedResult(hComm,&overlapRead,&nBytesRead,1)) {
						dwError = GetLastError();
						// normal result if not finished
						if(dwError == ERROR_IO_INCOMPLETE) continue;
						else {
							// an error occurred, try to recover
							ClearCommError (hComm, &dwError, &cs);
							break;
						}
					}
				}else {// some other error occurred 
					ClearCommError (hComm, &dwError, &cs);
					continue;
				}
		}
		if (nBytesRead) port->Parse(c);
	} while(1);
	/* clean out any pending bytes in the receive buffer */
	PurgeComm(hComm, PURGE_RXCLEAR);

	return 0L;
} /* end function (ReadThread) */


BOOL __declspec(dllexport) initcom(int port_num)
{
	char portname[5];

	//------------ (ALT) close comm first 
	CleanUp();

	//protocol class open
	port = new Port;
	
	//open comport
	if(port_num == 1) strcpy(portname,"COM1");
	if(port_num == 2) strcpy(portname,"COM2");

	// open the port here 
	if ((hComm=CreateFile (portname,
			GENERIC_READ|GENERIC_WRITE, 
			0, 
			NULL, 
			OPEN_EXISTING,
			FILE_FLAG_OVERLAPPED, //ALT
			NULL)) == INVALID_HANDLE_VALUE) {
		// handle error 
		PrintError("Failed to open port");
		return CleanUp();
	} /* end if (error opennig comport) */
	// set timeouts
	COMMTIMEOUTS cto = { MAXWORD, 0, 0, 5, 10 };
	DCB dcb;

	if(!SetCommTimeouts(hComm,&cto)) {
		PrintError("SetCommTimeouts failed");
		return CleanUp();
	}

	// -----------  Setup input/output buffers (ALT)
	if( SetupComm(hComm,2048,2048 ) == FALSE ) return CleanUp();

	// -----------  Clear port (ALT)
	PurgeComm(hComm,PURGE_TXABORT | PURGE_RXABORT |	PURGE_TXCLEAR | PURGE_RXCLEAR );

	//-------------- Set DCB (ALT)
    GetCommState(hComm,&dcb ); // Get comm's state first 
							   // too fill all the fields

	//memset(&dcb,0,sizeof(dcb)); //--do not need (ALT)
	//dcb.DCBlength = sizeof(dcb);

	dcb.BaudRate = 2400;
	dcb.fBinary = 1;
	dcb.fDtrControl = DTR_CONTROL_ENABLE;
	dcb.fOutxCtsFlow = 1;
	dcb.fRtsControl = DTR_CONTROL_HANDSHAKE;
	dcb.Parity = NOPARITY;
	dcb.StopBits = ONESTOPBIT;
	dcb.ByteSize = 8;
	if(!SetCommState(hComm,&dcb)){
		PrintError("SetCommState failed");
		return CleanUp();
	}

	//------------- setup overlapped stecture (ALT)

	overlapRead.Offset = 0 ;
    overlapRead.OffsetHigh = 0 ;
	overlapRead.hEvent = CreateEvent( NULL,    // no security
                                  TRUE,    // explicit reset req
                                  FALSE,   // initial event reset
                                  NULL ) ; // no name

	overlapWrite.Offset = 0 ;
    overlapWrite.OffsetHigh = 0 ;
	overlapWrite.hEvent = CreateEvent( NULL,    // no security
                                  TRUE,    // explicit reset req
                                  FALSE,   // initial event reset
                                  NULL ) ; // no name


//	EscapeCommFunction(hComm, SETDTR|SETRTS);

	/* create the thread for reading bytes */
	bReading = TRUE;
	if ((hThread=CreateThread (NULL /*def security */, 0 /* def stack size */, 
		(LPTHREAD_START_ROUTINE)ReadThread, NULL /* param to pass to thread */, 
			0, &threadID)) == INVALID_HANDLE_VALUE) {
		PrintError ("Error creating READ thread");
		return CleanUp();
	} // end if (error creating read thread)

	return TRUE;
}


//---------------------------- Overlapped I/O (ALT) ---------------------------
BOOL sendpacket1(unsigned char *packet)
{
	DWORD   write;
	DWORD	dwError;
	COMSTAT	cs;

	//------------- 1000 attemps --------------------
	int		attemps=3;

	while (attemps--) {
		ClearCommError (hComm, &dwError, &cs);
		BOOL success = WriteFile(hComm, packet, 1, &write, &overlapWrite);
		if (success) return TRUE;
		if (GetLastError()==ERROR_IO_PENDING) {
			//------- wait until sended -------------------
			while(!GetOverlappedResult(hComm,&overlapWrite,&write,1)) {
				dwError = GetLastError();
				// normal result if not finished
				if(dwError == ERROR_IO_INCOMPLETE) continue;
				else {
				   // an error occurred, try to recover
				   ClearCommError (hComm, &dwError, &cs);
				   //write=0;
				   break;
				}
			}
		}else {
				// some other error occurred
				// an error occurred, try to recover
				ClearCommError (hComm, &dwError, &cs);
		}
		if (write) return TRUE;
	}
	return FALSE;
}
BOOL sendpacket(unsigned char *packet)
{
	
	sendpacket1(packet);
	sendpacket1(packet+1);
	sendpacket1(packet+2);
	return sendpacket1(packet+3);
	
}

BOOL __declspec(dllexport) newgame(int color)
{
	//------- ALT repeat new game 4 times

	UINT i=4; // blacks tries 4 times
	if (color==2) i=1; // whites is just waiting.

	UINT timeout=TIMEOUT/i;
	while (i--) {
		DWORD StartCount;
		port->NewGame(color);
		StartCount = GetTickCount();
		do {
			if((GetTickCount()-StartCount) > timeout) break;
		} while(!(port->CheckOK()));
		if (port->CheckOK()) return TRUE;
	}
	return FALSE;
}

BOOL __declspec(dllexport) sendxy(int x, int y)
{
	DWORD StartCount;
	port->SendMove(x,y);
	StartCount = GetTickCount();
	do {
		if((GetTickCount()-StartCount) > TIMEOUT)
			return FALSE;
	}while(!(port->CheckOK()));
	return TRUE;
}

BOOL __declspec(dllexport) receivexy(int *px, int *py)
{
	BOOL ret;
	ret = port->RecieveMove(px,py);
	if(ret) port->ClearRecieve();
	return ret;
}


BOOL __declspec(dllexport) closecom(void)
{
	CleanUp();
	return TRUE;
}


